#!/usr/bin/env python

# Copyright (C) 2002 Jrg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import shelve
import threading

import item
import events

from helper import debug

#
# ReadWriteLock from Python Cookbook Recipe 6.3
#

class ReadWriteLock:
    def __init__(self):
        self._read_ready = threading.Condition(threading.Lock())
        self._readers = 0

    def acquire_read(self):
        # debug("ReadWriteLock: acquire read\n")
        self._read_ready.acquire()
        try:
            self._readers += 1
            # debug("ReadWriteLock: %d readers\n" % self._readers)
        finally:
            self._read_ready.release()

    def release_read(self):
        # debug("ReadWriteLock: release read\n")
        self._read_ready.acquire()
        try:
            self._readers -= 1
            # debug("ReadWriteLock: %d readers\n" % self._readers)
            if not self._readers:
                self._read_ready.notifyAll()
        finally:
            self._read_ready.release()

    def acquire_write(self):
        # debug("ReadWriteLock: acquire write\n")
        self._read_ready.acquire()
        while self._readers>0:
            self._read_ready.wait()

    def release_write(self):
        # debug("ReadWriteLock: release write\n")
        self._read_ready.release()

#
# songdb class
#

class songdb:
    def __init__(self, channel, songdbbase=None):
        self.songdbbase = songdbbase
        
        if self.songdbbase:
            self.songs = shelve.open(self.songdbbase + "_songs.db")
            self.artists = shelve.open(self.songdbbase + "_artists.db")
            
            debug("songdb: %d songs and %d artists registered\n" %
                  (len(self.songs), len(self.artists)))
        else:
            self.songs = {}
            self.artists = {}

        # empty cache of all artists and all songs
        # This variables contain tuples acting as caches for
        # all artists and all songs, respectively, in the database
        self.allartists = self.allsongs = None

        # this lock protects write access to the database
        self.rwlock = ReadWriteLock()

        channel.subscribe(events.songupdate, self.songupdate)

    def close(self):
        if self.songdbbase:
            self.songs.close()
            self.artists.close()

    def queryregister(self, path):
        """get song info from database or insert new one"""

        self.rwlock.acquire_write()
        
        if self.songs.has_key(path):
            song = self.songs[path]
        else:
            debug("songdb: new song: %s\n" % path)
            song = self.songs[path] = item.song(path)
            self.allsongs = None

        # insert song in artist and album database, if neccesary
        if not self.artists.has_key(song.artist):
            debug("songdb: new artist: %s\n" % song.artist)
            newartist = item.artist(song.artist)
            self.artists[song.artist] = newartist
            self.allartists = None
            events.hub.notify(events.artistaddedtosongdb(newartist))
            
        if not self.artists[song.artist].albums.has_key(song.album):
            artist = self.artists[song.artist]
            newalbum = item.album(song.artist, song.album)
            artist.albums[song.album] = newalbum
            self.artists[song.artist] = artist

            events.hub.notify(events.albumaddedtosongdb(newalbum))
            
        if path not in self.artists[song.artist].albums[song.album].songs:
            artist = self.artists[song.artist]
            artist.albums[song.album].songs.append(path)
            self.artists[song.artist] = artist
            
            # note that for performance reasons, we don't send notifications
            # of song additions

        self.rwlock.release_write()
            
        return song

    def updatesong(self, song):
        """updates entry of given song"""
        self.rwlock.acquire_write()
        self.songs[song.path] = song
        self.rwlock.release_write()
        
    def getsong(self, path):
        """returns song entry with given path"""
        self.rwlock.acquire_read()
        song = self.songs.get(path)
        self.rwlock.release_read()
        return song

    def getartists(self):
        """return all stored artists"""
        self.rwlock.acquire_read()
        # use cached value if existent
        if self.allartists:
            artists = self.allartists[:]
            self.rwlock.release_read()
            return artists
        else:
            artists = map(self.artists.get, self.artists.keys())
            self.rwlock.release_read()
            self.rwlock.acquire_write()
            self.allartists = artists[:]
            self.rwlock.release_write()
            return artists

    def getalbums(self, artist=None):
        """return albums of a given artist.

        artist has to be a string. If it is none, all stored
        albums are returned
        """
        if artist is None:
            albums = []
            for artist in self.getartists():
                for album in self.getalbums(artist.name):
                    albums.append(album)
            return albums
        else:
            self.rwlock.acquire_read()
            albums = self.artists[artist].albums.values()
            self.rwlock.release_read()
            return albums

    def getsongs(self, artist=None, album=None):
        """ returns song of given artist and given album
        
        artist and album are either strings or None
        If they are not given all artists and/or albums are chosen
        """
        
        if artist is None and album is None:
            # return all songs in songdb
            self.rwlock.acquire_read()
            # use cached value if existent
            if self.allsongs:
                songs = self.allsongs[:]
                self.rwlock.release_read()
                return songs
            else:
                songs = map(self.getsong, self.songs.keys())
                self.rwlock.release_read()
                self.rwlock.acquire_write()
                self.allsongs = songs[:]
                self.rwlock.release_write()
                return songs
        elif album is None:
            # return all songs of a given artist
            songs = []
            for album in self.getalbums(artist):
                for song in self.getsongs(artist, album.name):
                    songs.append(song)
            return songs
        elif artist is None:
            # return all songs which are on albums with the given name
            return [ aalbum for aalbum in self.getalbums()
                            if aalbum.name==album ]
        else:
            # return all songs on a given album of a given artist
            self.rwlock.acquire_read()
            songs = map(self.getsong,
                       self.artists[artist].albums[album].songs)
            self.rwlock.release_read()
            return songs
            
    def registerdirtree(self, dir):
        for name in os.listdir(dir):
                path = os.path.join(dir, name)
                if os.path.isdir(path):
                    try: self.registerdirtree(path)
                    except (IOError, OSError): pass
                elif name.endswith(".mp3") and os.access(path, os.R_OK):
                    try: self.queryregister(path)
                    except (IOError, OSError): pass

    # event handler

    def songupdate(self, event):
        self.updatesong(event.song)

#
# thread for automatic registering of songs in database
#

class songautoregisterer(threading.Thread):
    def __init__(self, db, basedir):
        self.db = db
        self.basedir = basedir

        threading.Thread.__init__(self)
        self.setDaemon(1)

    def run(self):
        self.db.registerdirtree(self.basedir)
        debug("songautoregisterer: exit\n")
